﻿//*****************************************************************************
//
//  By The Dummy Programmer
//  https://www.thedummyprogrammer.com
//
//*****************************************************************************

using System;
using System.ComponentModel;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;
using TDP.BaseServices.Infrastructure.Configuration;
using TDP.BaseServices.Infrastructure.Logging;
using TDP.BaseServices.Infrastructure.Logging.Abstract;
using TDP.BaseServices.Infrastructure.Net;
using TDP.BaseServices.Infrastructure.Security;
using TDP.PushLib.Client.Infrastructure;
using TDP.PushLib.Client.Infrastructure.Configuration;
using TDP.PushLib.Messages;
using TDP.PushLib.Messages.Core;

namespace TDP.PushLib.Client
{
    public class PushClient<T>
    {
        #region "Event definitions"
        public event EventHandler RegisterClientStarted;
        protected virtual void OnRegisterClientStarted(EventArgs e)
        {
            EventHandler handler = RegisterClientStarted;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public delegate void RegisterClientEndedDelegate(object sender, RegisterClientEndedEventArgs<T> e);
        public event RegisterClientEndedDelegate RegisterClientEnded;
        protected virtual void OnRegisterClientEnded(RegisterClientEndedEventArgs<T> e)
        {
            RegisterClientEndedDelegate handler = RegisterClientEnded;
            if (handler != null)
            {
                foreach (RegisterClientEndedDelegate singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public delegate void PushMessageReceivedDelegate(object sender, PushMessageReceivedEventArgs e);
        public event PushMessageReceivedDelegate PushMessageReceived;
        protected virtual void OnPushMessageReceived(PushMessageReceivedEventArgs e)
        {
            PushMessageReceivedDelegate handler = PushMessageReceived;
            if (handler != null)
            {
                foreach (PushMessageReceivedDelegate singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }

        public event EventHandler ConnectionClosed;
        protected virtual void OnConnectionClosed(EventArgs e)
        {
            EventHandler handler = ConnectionClosed;
            if (handler != null)
            {
                foreach (EventHandler singleCast in handler.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = singleCast.Target as ISynchronizeInvoke;
                    if ((syncInvoke != null) && (syncInvoke.InvokeRequired))
                        syncInvoke.Invoke(singleCast, new object[] { this, e });
                    else
                        singleCast(this, e);
                }
            }
        }
        #endregion
        private static ILogger _Log = new Logger();
        private bool _ClientRunning;
        private Socket _SocketClient;
        private Stream _SocketStream;
        private AuthenticationToken<T> _AuthenticationToken;

        private bool ValidateServerCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
        {
            ConfigReader CR = new ConfigReader();
            if (CR.Get(ConfigReaderKeys.KeyIgnoreSslErrors, Constants.DefaultIgnoreSslErrors))
                return true;

            if (sslPolicyErrors == SslPolicyErrors.None)
                return true;

            _Log.Info($"ChatClient -> ValidateServerCertificate() -> certificate error: {sslPolicyErrors}");

            return false;
        }

        public async void ConnectAsync(AuthenticationToken<T> authToken)
        {
            await Task.Run(() =>
            {
                _Log.Info($"PushLibClient -> ConnectAsync() -> async task started");

                _AuthenticationToken = authToken;
                RegisterClientResponse RegisterClientRsp;

                try
                {
                    EventArgs OnRegisterClientEA = new EventArgs();
                    OnRegisterClientStarted(OnRegisterClientEA);

                    ConfigReader CR = new ConfigReader();
                    _SocketClient = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                    IPAddress MyIPAddress = IPAddress.Parse(CR.Get(ConfigReaderKeys.KeyConnectIPAddress, Constants.DefaultConnectIPAddress));
                    IPEndPoint MyEndPoint = new IPEndPoint(MyIPAddress, CR.Get(ConfigReaderKeys.KeyConnectPort, Constants.DefaultConnectPort));
                    _SocketClient.Connect(MyEndPoint);

                    SslStream SslSocketStream;
                    if (!CR.Get(ConfigReaderKeys.KeyUseSsl, Constants.DefaultUseSsl))
                    {
                        _SocketStream = new NetworkStream(_SocketClient);
                    }
                    else
                    {
                        SslSocketStream = new SslStream(new NetworkStream(_SocketClient), false, new RemoteCertificateValidationCallback(ValidateServerCertificate));
                        SslSocketStream.AuthenticateAsClient(CR.Get(ConfigReaderKeys.KeySslTargetHost, Constants.DefaultSslTargetHost));
                        _SocketStream = SslSocketStream;
                    }

                    RegisterClient<T> RegisterClientMsg = new RegisterClient<T>();
                    RegisterClientMsg.AuthToken = _AuthenticationToken;
                    SocketHelper.SendObject(_SocketStream, RegisterClientMsg);

                    //TODO: Timeout for response?
                    //TODO: Add messages history to response?
                    RegisterClientRsp = (RegisterClientResponse)SocketHelper.ReceiveObject(_SocketStream);

                    if (RegisterClientRsp.Success)
                    {
                        _ClientRunning = true;
                        ManagePushMessagesAsync();
                    }
                }
                catch (Exception ex)
                {
                    _Log.Error($"PushLibClient -> ConnectAsync() -> an error occurred while connecting to the server", ex);

                    RegisterClientRsp = new RegisterClientResponse();
                    RegisterClientRsp.ErrorCode = -1;
                    RegisterClientRsp.Success = false;
                    RegisterClientRsp.Message = "Unable to connect to the server.";
                }
                
                RegisterClientEndedEventArgs<T> OnRegisterClientEndedEA = new RegisterClientEndedEventArgs<T>(RegisterClientRsp);
                OnRegisterClientEnded(OnRegisterClientEndedEA);

                _Log.Info($"PushLibClient -> ConnectAsync() -> async task ending");

                return;
            });
        }

        public void Disconnect()
        {
            try
            {
                if (_SocketClient.Connected)
                    _SocketClient.Close();
                _SocketClient.Dispose();

                EventArgs OnConnectionClosedEA = new EventArgs();
                OnConnectionClosed(OnConnectionClosedEA);
            }
            catch (Exception ex)
            {
                _Log.Error($"PushLibClient -> Disconnect() -> an error occurred", ex);
            }    
        }

        public async void ManagePushMessagesAsync()
        {
            await Task.Run(() =>
            {
                _Log.Info("PushLibClient -> ManagePushMessagesAsync() -> async task started");

                bool ConnectionBroken = false;

                try
                {
                    while (_ClientRunning)
                    {
                        PushMessage pushMessage = SocketHelper.ReceiveObject(_SocketStream) as PushMessage;

                        if (pushMessage == null)
                        {
                            _Log.Info("PushLibClient -> ManagePushMessagesAsync() -> received a null push message, closing connection");

                            // Connection has been broken
                            _ClientRunning = false;
                            ConnectionBroken = true;
                            break;
                        }

                        _Log.Info("PushLibClient -> ManagePushMessagesAsync() -> received push message, raising event");

                        PushMessageReceivedEventArgs e = new PushMessageReceivedEventArgs(pushMessage);
                        OnPushMessageReceived(e);
                    }
                }
                catch (ObjectDisposedException ex)
                {
                    ConnectionBroken = true;
                    _Log.Error("PushLibClient -> ManagePushMessagesAsync() -> connection closed", ex);
                }
                catch (Exception ex)
                {
                    _Log.Error("PushLibClient -> ManagePushMessagesAsync() -> connection closed", ex);
                }

                if (ConnectionBroken)
                {
                    EventArgs OnConnectionClosedEA = new EventArgs();
                    OnConnectionClosed(OnConnectionClosedEA);
                }

                _Log.Info("PushLibClient -> ManagePushMessagesAsync() -> async task ending");
            });
        }


    }
}
